package telinc.telicraft.api;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;

import com.google.common.collect.Lists;
import com.google.common.collect.Sets;

import net.minecraft.block.Block;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.inventory.InventoryCrafting;
import net.minecraft.item.Item;
import net.minecraft.item.ItemStack;
import net.minecraft.item.crafting.IRecipe;

import cpw.mods.fml.common.FMLLog;
import cpw.mods.fml.common.Loader;

import telinc.telicraft.TelicraftMain;

public final class TelicraftAPI {
    private static List<ISharpenerHandler> sharpenerHandlers = Lists.newArrayList();
    private static List<IAdvancedCraftingTableHandler> actHandlers = Lists.newArrayList();
    private static Set<Integer> bannedBlocksPetrify = Sets.newHashSet();
    private static Set<Block> meteorBombImmuneBlocks = Sets.newHashSet();

    /**
     * Adds a sharpening recipe to the Sharpener.
     * 
     * @param par1 The input item.
     * @param par2ItemStack The output item.
     * @param par3 The experience points you gain.
     */
    public static void addSharpenerRecipe(int par1, ItemStack par2ItemStack, float par3) {
        try{
            if(isTelicraftLoaded()){
                telinc.telicraft.item.crafting.SharpenerRecipes.getInstance().addRecipe(par1, par2ItemStack, par3);
            }else{
                FMLLog.finer("[Telicraft API] Telicraft not detected, ignoring new Sharpener recipe request.", new Object[0]);
                return;
            }
        }catch(Exception e){
            e.printStackTrace();
        }
    }

    /**
     * Adds a sharpening recipe to the Sharpener.
     * 
     * @param par1 The input item.
     * @param par2 The metadata of the input.
     * @param par3ItemStack The output item.
     * @param par4 The experience points you gain.
     */
    public static void addSharpenerRecipe(int par1, int par2, ItemStack par3ItemStack, float par4) {
        try{
            if(isTelicraftLoaded()){
                telinc.telicraft.item.crafting.SharpenerRecipes.getInstance().addRecipe(par1, par2, par3ItemStack, par4);
            }else{
                FMLLog.finer("[Telicraft API] Telicraft not detected, ignoring new Sharpener recipe request.", new Object[0]);
                return;
            }
        }catch(Exception e){
            e.printStackTrace();
        }
    }

    public static void addACTRecipe(ItemStack output, Object... params) {
        try{
            if(isTelicraftLoaded()){
                telinc.telicraft.item.crafting.ACTManager.getInstance().addRecipe(output, params);
            }else{
                FMLLog.finer("[Telicraft API] Telicraft not detected, ignoring new Advanced Crafting Table recipe request.", new Object[0]);
                return;
            }
        }catch(Exception e){
            e.printStackTrace();
        }
    }

    public static void addACTRecipe(IRecipe recipe) {
        try{
            if(isTelicraftLoaded()){
                telinc.telicraft.item.crafting.ACTManager.getInstance().getRecipeList().add(recipe);
            }else{
                FMLLog.finer("[Telicraft API] Telicraft not detected, ignoring new Advanced Crafting Table recipe request.", new Object[0]);
                return;
            }
        }catch(Exception e){
            e.printStackTrace();
        }
    }

    public static void addShapelessACTRecipe(ItemStack output, Object... params) {
        try{
            if(isTelicraftLoaded()){
                telinc.telicraft.item.crafting.ACTManager.getInstance().addShapelessRecipe(output, params);
            }else{
                FMLLog.finer("[Telicraft API] Telicraft not detected, ignoring new Advanced Crafting Table recipe request.", new Object[0]);
                return;
            }
        }catch(Exception e){
            e.printStackTrace();
        }
    }

    /**
     * Adds a banned block to the Direct Petrifying Potion's list.<br />
     * Banned blocks will not turn into stone when hit by the potion.
     * 
     * @param par1Block The block to add. It will be automatically converted to its ID.
     */
    public static void addBannedBlockForPetrify(Block par1Block) {
        bannedBlocksPetrify.add(par1Block.blockID);
    }

    /**
     * Adds a banned block ID to the Direct Petrifying Potion's list.<br />
     * Banned blocks will not turn into stone when hit by the potion.
     * 
     * @param par1 The block ID to add.
     */
    public static void addBannedBlockForPetrify(int par1) {
        bannedBlocksPetrify.add(par1);
    }

    /**
     * Makes a block immune to the Meteor Bomb.
     * 
     * @param par1Block The block to add.
     */
    public static void addMeteorBombImmuneBlock(Block par1Block) {
        meteorBombImmuneBlocks.add(par1Block);
    }

    /**
     * Makes a block immune to the Meteor Bomb.
     * 
     * @param par1 The block's ID. It will use Block.blocksList to figure out which block it is.
     * @deprecated Please use the version of this that adds a block directly and not by its ID when possible.
     */
    @Deprecated
    public static void addMeteorBombImmuneBlock(int par1) {
        meteorBombImmuneBlocks.add(Block.blocksList[par1]);
    }

    /**
     * Returns the HashSet of the blocks immune to the Meteor Bomb.
     */
    public static Set<Block> getMeteorBombImmuneBlocks() {
        return meteorBombImmuneBlocks;
    }

    /**
     * Returns the HashSet of the blocks banned from being petrified.
     */
    public static Set<Integer> getBannedBlocksPetrify() {
        return bannedBlocksPetrify;
    }

    /**
     * Used to get a block from Telicraft.
     * 
     * @param internalName The field name. Refer to {@link TelicraftMain} for the names.
     * @return A block. If it can't find anything, it will return stone.
     */
    public static Object getBlock(String internalName) {
        try{
            Class<?> telicraftClass = Class.forName("telinc.telicraft.TelicraftMain");
            Field objectToReturn = telicraftClass.getDeclaredField(internalName);

            if(objectToReturn.getType() == Block.class){
                return objectToReturn.get(telicraftClass);
            }else{
                FMLLog.warning("[Telicraft API] Tried to access a non-block field through getBlock(String). This is a serious programming mistake, please consider talking to the mod author of the mod that uses this API.", new Object[0]);
                return Block.stone;
            }
        }catch(ClassNotFoundException e){
            FMLLog.finer("[Telicraft API] Could not find class TelicraftMain, mod is probably not installed.", e);
            return Block.stone;
        }catch(NoSuchFieldException e){
            FMLLog.warning("[Telicraft API] Could not find requested block. Please ensure that the internal name is the correct variable name.", e);
            return Block.stone;
        }catch(SecurityException e){
            FMLLog.warning("[Telicraft API] Security excepetion occured. Please talk to the mod author using this API about the problem.", e);
        }catch(IllegalArgumentException e){
            FMLLog.warning("[Telicraft API] Method getBlock(String) threw an IllegalArgumentExcpetion. This should not happen unless somebody's using a modified version of the Telicraft API.", e);
        }catch(IllegalAccessException e){
            FMLLog.warning("[Telicraft API] Tried to access protected or private member of TelicraftMain. Bug?", e);
        }

        return Block.stone;
    }

    /**
     * Used to get a block from Telicraft.
     * 
     * @param internalName The field name. Refer to {@link TelicraftMain} for the names.
     * @return An item. If the method can't find anything, it will return a stick.
     */
    public static Object getItem(String internalName) {
        try{
            Class<?> telicraftClass = Class.forName("telinc.telicraft.TelicraftMain");
            Field objectToReturn = telicraftClass.getDeclaredField(internalName);

            if(objectToReturn.getType() == Item.class){
                return objectToReturn.get(telicraftClass);
            }else{
                FMLLog.warning("[Telicraft API] Tried to access a non-item field through getItem(String). This is a serious programming mistake, please consider talking to the mod author of the mod that uses this API.", new Object[0]);
                return Item.stick;
            }
        }catch(ClassNotFoundException e){
            FMLLog.finer("[Telicraft API] Could not find class TelicraftMain, mod is probably not installed.", e);
            return Block.stone;
        }catch(NoSuchFieldException e){
            FMLLog.warning("[Telicraft API] Could not find requested item. Please ensure that the internal name is the correct variable name.", e);
            return Block.stone;
        }catch(SecurityException e){
            FMLLog.warning("[Telicraft API] Security excepetion occured. Please talk to the mod author using this API about the problem.", e);
        }catch(IllegalArgumentException e){
            FMLLog.warning("[Telicraft API] Method getItem(String) threw an IllegalArgumentExcpetion. This should not happen unless somebody's using a modified version of the Telicraft API.", e);
        }catch(IllegalAccessException e){
            FMLLog.warning("[Telicraft API] Tried to access protected or private member of TelicraftMain. Bug?", e);
        }

        return Item.stick;
    }

    /**
     * Uses reflection to try and print out a message to Telicraft's logger.
     * 
     * @param lvl The level of the message.
     * @param message The message itself.
     * @param object Any kind of object. If you want to print just a message without an object information, see {@link #printMessageToLogger(Level, String)}
     */
    private static void printMessageToLogger(Level lvl, String message, Object object) {
        try{
            Class<?> telicraftClass = Class.forName("telinc.telicraft.TelicraftMain");
            Class<?> loggerClass = Class.forName("java.util.logging.Logger");
            Field logger = telicraftClass.getDeclaredField("tcLog");
            Method logMethod = logger.getDeclaringClass().getDeclaredMethod("log", Level.class, String.class, Object.class);
            logMethod.invoke(loggerClass.newInstance(), lvl, message, object);
        }catch(ClassNotFoundException ex){
            FMLLog.finer("[Telicraft API] Could not find class TelicraftMain, mod is probably not installed.", ex);
        }catch(NoSuchFieldException ex){
            FMLLog.warning("[Telicraft API] Found TelicraftMain.class, but did not find field 'tcLog'. This will not happen unless you are using a modified version of the Telicraft API.", ex);
        }catch(SecurityException ex){
            FMLLog.warning("[Telicraft API] Security excepetion occured. Please talk to the mod author using this API about the problem.", ex);
        }catch(NoSuchMethodException ex){
            FMLLog.warning("[Telicraft API] Could not find method 'log' in class Logger. This is a normal Java class, so you are probably using a modified/outdata Java version.", ex);
        }catch(IllegalAccessException ex){
            FMLLog.warning("[Telicraft API] Either field 'tcLog' or method 'log' are private or protected. Bug?", ex);
        }catch(IllegalArgumentException ex){
            FMLLog.warning("[Telicraft API] Method printMessageToLogger(Level, String, Object) threw an IllegalArgumentExcpetion. This should not happen unless somebody's using a modified version of the Telicraft API.", ex);
        }catch(InvocationTargetException ex){
            FMLLog.warning("[Telicraft API] Method 'log' from normal Java class Logger threw an exception. What?!", ex);
        }catch(InstantiationException ex){
            FMLLog.warning("[Telicraft API] Class java.util.logging.Logger is either abstract, an interface, an array class, or a primitive type. Modified Java?", ex);
        }
    }

    /**
     * Uses reflection to try and print out a message to Telicraft's logger. This one will not give an extra Throwable object to the method.<br />
     * If you want to give an extra object, see {@link #printMessageToLogger(Level, String, Object)}
     * 
     * @param lvl The level of the message.
     * @param message The message itself.
     */
    private static void printMessageToLogger(Level lvl, String message) {
        try{
            Class<?> telicraftClass = Class.forName("telinc.telicraft.TelicraftMain");
            Class<?> loggerClass = Class.forName("java.util.logging.Logger");
            Object loggerInstance = loggerClass.newInstance();
            Field logger = telicraftClass.getDeclaredField("tcLog");
            Method logMethod = logger.getDeclaringClass().getDeclaredMethod("log", Level.class, String.class);
            logMethod.invoke(loggerInstance, lvl, message);
        }catch(ClassNotFoundException ex){
            FMLLog.finer("[Telicraft API] Could not find class TelicraftMain, mod is probably not installed.", ex);
        }catch(NoSuchFieldException ex){
            FMLLog.warning("[Telicraft API] Found TelicraftMain.class, but did not find field 'tcLog'. This will not happen unless you are using a modified version of the Telicraft API.", ex);
        }catch(SecurityException ex){
            FMLLog.warning("[Telicraft API] Security excepetion occured. Please talk to the mod author using this API about the problem.", ex);
        }catch(NoSuchMethodException ex){
            FMLLog.warning("[Telicraft API] Could not find method 'log' in class Logger. This is a normal Java class, so you are probably using a modified/outdata Java version.", ex);
        }catch(IllegalAccessException ex){
            FMLLog.warning("[Telicraft API] Either field 'tcLog' or method 'log' are private or protected. Bug?", ex);
        }catch(IllegalArgumentException ex){
            FMLLog.warning("[Telicraft API] Method printMessageToLogger(Level, String, Throwable) threw an IllegalArgumentExcpetion. This should not happen unless somebody's using a modified version of the Telicraft API.", ex);
        }catch(InvocationTargetException ex){
            FMLLog.warning("[Telicraft API] Method 'log' from normal Java class Logger threw an exception. What?!", ex);
        }catch(InstantiationException ex){
            FMLLog.warning("[Telicraft API] Class java.util.logging.Logger is either abstract, an interface, an array class, or a primitive type. Modified Java?", ex);
        }
    }

    /**
     * Gets if Telicraft is currently loaded. Please use for recipes, you don't want to have Sticks instead of Adamant in your recipe.
     */
    public static boolean isTelicraftLoaded() {
        return Loader.isModLoaded("telicraft");
    }

    /**
     * Call to simulate an item being taken out of a Sharpener's output slot.
     * 
     * @param player The player instance.
     * @param item The item stack taken out.
     */
    public static void onItemSharpened(EntityPlayer player, ItemStack item) {
        for(ISharpenerHandler handler : sharpenerHandlers){
            handler.onItemTakenOut(player, item);
        }
    }

    /**
     * Call to simulate an item being taken out of an ACT's output slot.
     * 
     * @param par1EntityPlayer The player that took it out.
     * @param par2ItemStack The item.
     * @param par3InventoryCrafting The InventoryCrafting belonging to the table.
     */
    public static void onItemCraftedACT(EntityPlayer par1EntityPlayer, ItemStack par2ItemStack, InventoryCrafting par3InventoryCrafting) {
        for(IAdvancedCraftingTableHandler handler : actHandlers){
            handler.onItemCrafted(par1EntityPlayer, par2ItemStack, par3InventoryCrafting);
        }
    }

    /**
     * Register a sharpener handler.
     * 
     * @param handler An instance of ISharpenerHandler.
     */
    public static void registerSharpenerHandler(ISharpenerHandler handler) {
        sharpenerHandlers.add(handler);
    }

    /**
     * Register a sharpener handler.
     * 
     * @param handler An instance of IAdvancedCraftingTableHandler.
     */
    public static void registerACTHandler(IAdvancedCraftingTableHandler handler) {
        actHandlers.add(handler);
    }
}